#define NAME	 "Change"
#define VERSION  "2"
#define REVISION "37"

/* Possible define CLIONLY for Change.c and ChangeTexts.c to compile a CLI
   only version. Note: you must not link ChangeGUI.o to this version.
   The CLIONLY version of Change 2.29 is 5768 bytes, where normal version
   is 9356 bytes (3588 bytes shorter).

   This program is multi-reentrant and can be made resident.
*/

/* Programmheader

	Name:		Change
	Author:		SDI
	Distribution:	PD
	Description:	searches and changes characters in files
	Compileropts:	-
	Linkeropts:	-l amiga -gsi

 2.5   06.08.95 : fixed error in SDI_WriteBuf (now 1.4)
 2.6   15.10.95 : better return codes
 2.7   31.10.95 : fixed some bugs
 2.8   10.12.95 : fixed bug with PATTERN_BACK
 2.9   25.12.95 : better window flags, now also PubScreen useable
 2.10  31.12.95 : SDI_OpenBF function in includes new, a bit shorter,
 	SDI_FileReq instead of own ASL function, now English
 2.11  02.02.96 : fixed write error in texts
 2.12  04.02.96 : fixed bug with strings in GUI
 2.13  05.02.96 : reworked gadget handling, field of gadgets instead of
	search routine
 2.14  06.02.96 : fixed bug in includes, added TestOS
 2.15  10.02.96 : used own startup code --> TestOS removed again
 2.16  14.02.96 : bugfix - GUI was always started in last version
 2.17  31.03.96 : fixed error with POSITION option
 2.18  04.05.96 : fixed error with ReadArgs
 2.19  02.08.96 : on CTRL_C break and FROM == TO, source is saved
 2.20  05.11.96 : changed the end text, to print out a bit better info
 2.21  27.11.96 : OVERWRITE didn't work, removed bug with very short files
 2.22  18.03.97 : SDI includes changed a bit
 2.23  01.08.97 : some source changes, translated source to English
 2.24  24.12.97 : totally rewritten scan routines, lots of changes, shorter
 2.25  26.12.97 : bug fixes in scan routines
 2.26  29.12.97 : again bug-fixes, fixed errors with SAS compilation
 2.27  03.01.98 : startup code now internal, made SAS compilable (SMakeFile),
	removed GetFile gadget image
 2.28  04.01.98 : removed some obsolete stuff
 2.29  07.01.98 : hope I fixed bugs crashing the GUI, moved GUI to
	ChangeGUI.c completely
 2.30  02.07.98 : bug fix with patterns
 2.31  05.07.98 : bug fix with OVERWRITE
 2.32  03.12.98 : bug fix with POSITION and PATTERN
 2.33  15.12.98 : added pattern support
 2.34  22.12.98 : fixed bug introduced in last version
 2.35  31.12.98 : no more errors when overwriting source and
	using patterns
 2.36  01.01.00 : fixed problem with PATH option
 2.37  07.01.00 : added RFILE and SFILE options
*/
 
#include <proto/dos.h>
#include <proto/exec.h>
#include <exec/memory.h>
#include "SDI_ASM_STD_protos.h"
#include "SDI_defines.h"
#include "Change.h"
#include "ChangeTexts.h"
#include "ChangeFunc.h"

#ifndef CLIONLY
  #include "ChangeGUI.h"
#endif

struct DosLibrary *DOSBase = 0;
struct ExecBase *  SysBase = 0;

/* =============================== start =============================== */

LONG start(void)
{
  LONG ret = RETURN_FAIL;
  struct DosLibrary *dosbase;

  SysBase = (*((struct ExecBase **) 4));

  if((dosbase = (struct DosLibrary *) OpenLibrary("dos.library", 37)))
  {
    struct Process *task;

    DOSBase = dosbase;
    if(!(task = (struct Process *) FindTask(0))->pr_CLI)
    {
      struct Message *msg;
      WaitPort(&task->pr_MsgPort);
      msg = GetMsg(&task->pr_MsgPort);
#ifdef CLIONLY
      ret = RETURN_FAIL;
#else
      ret = DoGUI();
#endif
      Forbid();
      ReplyMsg(msg);
    }
#ifndef CLIONLY
    else if(task->pr_Arguments[1] == 0)
      ret = DoGUI();
#endif
    else
    {
      struct Args args;
      struct RDArgs *rda;

      SDI_memset(&args, 0, sizeof(struct Args));

      if((rda = (struct RDArgs *) AllocDosObject(DOS_RDARGS, 0)))
      {
        rda->RDA_ExtHelp = TEXT_CLI_HELP;
        if(ReadArgs(PARAM, (LONG *) &args, rda))
        {
          UWORD *sstr = 0, *rstr = 0;
          STRPTR str;
          ULONG err = 0;
          struct ChangeData cd;

          SDI_memset(&cd, 0, sizeof(struct ChangeData));

	  if(args.sfile)
	  {
	    if(!(sstr = LoadStringFile(args.sfile, args.hex, &str)))
	      err = 1;
	    else
	    {
              cd.l[0] = args.hex ? GetCharHEX(str, sstr) :
              GetCharSTRING(str, sstr);
              cd.string[0] = sstr;
            }
	    if(!args.to && args.sstring)
              args.to = args.sstring;
	  }
	  else if(args.sstring)
	  {
            cd.l[0] = args.hex ? GetCharHEX(args.sstring, cd.str[0]) :
            GetCharSTRING(args.sstring, cd.str[0]);
            cd.string[0] = cd.str[0];
          }

          if(!args.nochange)
          {
	    if(args.rfile)
	    {
	      if(!(rstr = LoadStringFile(args.rfile, args.hex, &str)))
	        err = 1;
	      else
	      {
                cd.l[1] = args.hex ? GetCharHEX(str, rstr) :
                GetCharSTRING(str, rstr);
                cd.string[1] = rstr;
              }
	    }
	    else
	    {
	      if(args.rstring)
                cd.l[1] = args.hex ? GetCharHEX(args.rstring, cd.str[1]) :
                GetCharSTRING(args.rstring, cd.str[1]);
              cd.string[1] = cd.str[1];
              /* cd.l[1] is 0, if no rstring! */
	    }
          }

          if(!cd.string[0] || !*cd.string[0] || !cd.l[0])
            SetIoErr(ERROR_REQUIRED_ARG_MISSING);
          else if(!err)
          {
	    ULONG h, i;

            if(args.found)		cd.flags |= CHANGEFLAG_FOUND;
            if(args.overwrite)		cd.flags |= CHANGEFLAG_OVERWRITE;
	    if(args.length)		cd.flags |= CHANGEFLAG_LENGTH;
	    if(args.nosize) 		cd.flags |= CHANGEFLAG_NOSIZE;
	    if(args.path)		cd.flags |= CHANGEFLAG_PATH;
	    if(args.position)		cd.flags |= CHANGEFLAG_POSITION;

	    if(args.nochange)		cd.flags |= CHANGEFLAG_NOCHANGE;
	    if(args.all)		cd.flags |= CHANGEFLAG_ALL;

            for(h = 0; h < (args.nochange ? 1 : 2); ++h)
            {
              PutStr(h == 0 ? TEXT_SEARCH_STR : TEXT_REPLACE_STR);
              for(i = 0; i < cd.l[h]; ++i)		/* printout strings */
              {
                if(cd.str[h][i] == PATTERN_BACK)
                  PutStr("\033[7m" PATTERN_BACK_OUT "\033[27m");
                else if(SDI_isprint(cd.string[h][i]))
                  FPutC(Output(), cd.string[h][i]);
                else
                  PutStr("\033[7m.\033[27m");
              }
              PutStr("'\n");
            }
            cd.position = args.position ? SDI_strtoul(args.position, 0, 16) : 0;
	    ret = DoDirectoryScan(args.from, args.to, &cd);
          }
	  if(sstr)
	    FreeVec(sstr);
	  if(rstr)
	    FreeVec(rstr);
          FreeArgs(rda);
        } /* ReadArgs */
        FreeDosObject(DOS_RDARGS, rda);
      } /* AllocDosObject */

      if(ret)
        PrintFault(IoErr(), 0);
    } /* else */
    CloseLibrary((struct Library *) DOSBase);
  } /* OpenLibrary dos.library */

  return ret;
}

/* This scans a directory and calls DoSearch/DoChange for every file. */
LONG DoDirectoryScan(STRPTR from, STRPTR to, struct ChangeData *cd)
{
  struct AnchorPath *APath;
  ULONG retval = 1;
  LONG ret = 0, onfile = 1;
  STRPTR dest = to;

  if(to && *to)
  {
    BPTR lock;
	      
    if((lock = Lock(to, SHARED_LOCK)))
   {
      struct FileInfoBlock *fib;
     if((fib = (struct FileInfoBlock *) AllocDosObject(DOS_FIB, 0)))
     {
        if(Examine(lock, fib))
        {
	  if(fib->fib_DirEntryType > 0)
	    onfile = 0;
	}
	FreeDosObject(DOS_FIB, fib);
      }
      UnLock(lock);
    }
  }

  if((APath = (struct AnchorPath *) AllocMem(sizeof(struct AnchorPath) +
  512, MEMF_PUBLIC|MEMF_CLEAR)))
  {
    STRPTR bufptr;
    APath->ap_Strlen = 256;

    bufptr = APath->ap_Buf+256;

    retval = MatchFirst(from, APath);
    while(!retval && !ret && !CTRL_C)
    {
      if(onfile == 2)
        retval = ERROR_BAD_TEMPLATE;
      else if(APath->ap_Flags & APF_DIDDIR)
        APath->ap_Flags &= ~APF_DIDDIR; /* clear flag */
      else if(APath->ap_Info.fib_DirEntryType > 0)
      {
        if(cd->flags & CHANGEFLAG_ALL)
            APath->ap_Flags |= APF_DODIR;
      }
      else
      {
	Printf("File: '%s': ", APath->ap_Buf);
	Flush(Output());
	if(cd->flags & CHANGEFLAG_NOCHANGE)
	  ret = DoSearch(APath->ap_Buf, cd->string[0], cd->l[0], cd->flags);
	else
	{
	  if(to && *to)
	  {
	    if(onfile)
	      ++onfile;
	    else
	    {
	      dest = bufptr;
	      CopyMem(to, bufptr, SDI_strlen(to)+1);
	      AddPart(bufptr, FilePart(APath->ap_Buf), 256);
	    }
	  }

	  ret = DoChange(APath->ap_Buf, cd->string[0], cd->l[0], dest, cd->string[1], cd->l[1],
	  cd->flags, cd->position);
	}
      }
      if(!retval && !ret)
        retval = MatchNext(APath);
    }
    MatchEnd(APath);

    if(retval == ERROR_NO_MORE_ENTRIES)
      retval = 0;

    FreeMem(APath, sizeof(struct AnchorPath) + 512);
  }

  if(CTRL_C)
  {
    Printf("\n");
    retval = ERROR_BREAK;
  }

  if(retval)
  {
    ret = RETURN_ERROR;
    SetIoErr(retval);
  }

  return ret;
}

UWORD *LoadStringFile(STRPTR name, ULONG hex, STRPTR *srcpos)
{
  struct FileInfoBlock *fib;
  STRPTR r = 0;
  BPTR fh;

  if((fib = (struct FileInfoBlock *) AllocDosObject(DOS_FIB, 0)))
  {
    if((fh = Open(name, MODE_OLDFILE)))
    {
      if(ExamineFH(fh, fib))
      {
        if((r = (STRPTR) AllocVec(fib->fib_Size*3+1, MEMF_CLEAR|MEMF_ANY)))
        {
          if(Read(fh, r+(fib->fib_Size<<1),fib->fib_Size) == fib->fib_Size)
            *srcpos = r+(fib->fib_Size<<1);
          else
          {
            FreeVec(r); r = 0;
          } 
        }
      }
      Close(fh);
    }
    FreeDosObject(DOS_FIB, fib);
  }
  return (UWORD *) r;
}

